/*
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */


#import <Cocoa/Cocoa.h>
#import "IORegOutlineViewNode.h"
#import "IORegDetailOutlineViewNode.h"
#import "IORegInfoGatherer.h"
#import "OutlineViewAdditions.h"

@interface IORegistryController : NSObject <IORegGathererListener>
{
    IBOutlet id IORegDetailedOutput;
    IBOutlet id IORegDetailedOutputDrawer;
    IBOutlet id IORegOutputOV;
    IBOutlet id PlanePopUpButton;
    IBOutlet id RefreshButton;
    IBOutlet id RefreshCheckBox;
    
    IORegOutlineViewNode *  _rootNode;
    IORegInfoGatherer *     _infoGatherer;
    
    IORegDetailOutlineViewNode *       _detailRootNode;
}

- (IBAction)Refresh:(id)sender;
- (IBAction)SaveOutput:(id)sender;
- (IBAction)ToggleAutoRefresh:(id)sender;
- (IBAction)TogglePlane:(id)sender;

- (void)ioRegInfoGathererInformationDidChange:(IORegInfoGatherer *)aGatherer;
- (void)expandOutlineViewItems;

- (void)ioregItemSingleClicked:(id)sender;
- (void)ioregItemDoubleClicked:(id)sender;

- (void)populateNode:(IORegDetailOutlineViewNode *)node withKey:(NSString *)key value:(id)value;
- (void)populateNode:(IORegDetailOutlineViewNode *)node withContentsOfDictionary:(NSDictionary *)dictionary;
- (void)populateNode:(IORegDetailOutlineViewNode *)node withContentsOfArray:(NSArray *)array;

- (void)detailDrawerDidOpen:(NSNotification *)notification;

@end
