/*
 * Copyright (c) 2004 Apple Computer, Inc. All rights reserved.
 *
 * @APPLE_LICENSE_HEADER_START@
 * 
 * This file contains Original Code and/or Modifications of Original Code
 * as defined in and that are subject to the Apple Public Source License
 * Version 2.0 (the 'License'). You may not use this file except in
 * compliance with the License. Please obtain a copy of the License at
 * http://www.opensource.apple.com/apsl/ and read it before using this
 * file.
 * 
 * The Original Code and all software distributed under the License are
 * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
 * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
 * Please see the License for the specific language governing rights and
 * limitations under the License.
 * 
 * @APPLE_LICENSE_HEADER_END@
 */


#include "AppleUSBUHCI.h"


// method in 1.8 and 1.8.1
IOReturn
AppleUSBUHCI::UIMCreateInterruptTransfer(
                                         short				functionNumber,
                                         short				endpointNumber,
                                         IOUSBCompletion                completion,
                                         IOMemoryDescriptor *		CBP,
                                         bool				bufferRounding,
                                         UInt32				bufferSize,
                                         short				direction)
{
    // deprecated
    return kIOReturnBadArgument;
}

// method in 1.8 and 1.8.1
IOReturn
AppleUSBUHCI::UIMCreateBulkTransfer(
                                    short				functionNumber,
                                    short				endpointNumber,
                                    IOUSBCompletion			completion,
                                    IOMemoryDescriptor *		CBP,
                                    bool				bufferRounding,
                                    UInt32				bufferSize,
                                    short				direction)
{
    // deprecated
    return kIOReturnBadArgument;
}

IOReturn
AppleUSBUHCI::UIMCreateControlTransfer(
                                       short				functionNumber,
                                       short				endpointNumber,
                                       IOUSBCommand*			command,
                                       void *				CBP,
                                       bool				bufferRounding,
                                       UInt32				bufferSize,
                                       short				direction)
{
    // deprecated
    return kIOReturnBadArgument;
}

// method in 1.8 and 1.8.1
IOReturn
AppleUSBUHCI::UIMCreateControlTransfer(
                                       short				functionNumber,
                                       short				endpointNumber,
                                       IOUSBCompletion			completion,
                                       IOMemoryDescriptor *		CBP,
                                       bool				bufferRounding,
                                       UInt32				bufferSize,
                                       short				direction)
{
    // deprecated
    return kIOReturnBadArgument;
}

// method in 1.8 and 1.8.1
IOReturn
AppleUSBUHCI::UIMCreateControlTransfer(
                                       short				functionNumber,
                                       short				endpointNumber,
                                       IOUSBCompletion			completion,
                                       void *				CBP,
                                       bool				bufferRounding,
                                       UInt32				bufferSize,
                                       short				direction)
{
    // deprecated
    return kIOReturnIPCError;
}



